#!/usr/bin/env python
#
#  pyHatching -- Apply a hatching effect to an image.
#  Copyright (C) 2001 Fabio Dias Moreira
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
## Future directives #################################################

from __future__ import nested_scopes
from __future__ import generators
from __future__ import division

## Module description ################################################

"""imageHandling.py -- Useful image operations.

showArray and showSurface adapted from pyGame's surfarray example (www.pygame.org) 
"""

## Imports ###########################################################

from math import *

## pyGame -- just import things necessary for image handling
import pygame
import pygame.image
import Numeric as N
import pygame.surfarray as surfarray
from pygame.locals import *

## Constants #########################################################

## Globals ###########################################################

## Classes ###########################################################

## Functions #########################################################

def nudgeImage(image, dx, dy):
    imageWidth, imageHeight = image.shape
    n_dx = dx % imageWidth
    n_dy = dy % imageHeight
    base = N.zeros(image.shape, N.Float)
    if n_dx == 0:
        if n_dy == 0:
            return image
        base[:, :dy] = image[:,-dy:]
        base[:, dy:] = image[:,:-dy]
        return base
    if n_dy == 0:
        base[:dx, :] = image[-dx:, :]
        base[dx:, :] = image[:-dx, :]
        return base
    base[:dx, :dy] = image[-dx:,-dy:]
    base[dx:, :dy] = image[:-dx,-dy:]
    base[:dx, dy:] = image[-dx:,:-dy]
    base[dx:, dy:] = image[:-dx,:-dy]
    return base

def circularConvolution(image, filter, hotspot=None):
    filterWidth, filterHeight = filter.shape[:2]
    if hotspot is None:
        hotspot = (filterWidth//2, filterHeight//2)
    base = N.zeros(image.shape, N.Float)
    for i in range(filterWidth):
        for j in range(filterHeight):
            base += filter[i, j] * \
                    nudgeImage(image, i - hotspot[0], j - hotspot[1])
    return base

def normalizeImage(image):
    return N.minimum(N.maximum(image, \
                               N.zeros(image.shape, N.Float)), \
                     N.ones(image.shape, N.Float))

def blurImage(image, order = 2):
    size = 2 * order - 1
    filter = N.fromfunction(lambda x, y: N.cos((x - (order - 1)) * (pi/(2*size))) * N.cos((y - (order - 1)) * (pi/(2*size))), (size, size))
    filter *= 1/N.sum(N.sum(filter))
    return circularConvolution(image, filter, (order-1, order-1))
    
def showArray(array, name='image', multiplier=1):
    screen = pygame.display.set_mode(array.shape, 0, 32)
    # Normalize image, convert to grayscale
    surfarray.blit_array(screen, (N.absolute(array) * 255 * multiplier).astype(N.UnsignedInt8) * 65793)
    pygame.display.flip()
    pygame.display.set_caption(name)
    while 1:
        e = pygame.event.poll()
        if e.type == MOUSEBUTTONDOWN: break
        elif e.type == KEYDOWN and e.key == K_s:
            pygame.image.save(screen, name + '.bmp')
        elif e.type == QUIT: raise SystemExit

def showSurface(surface, name='image'):
    showArray(surfaceToArray(surface), name)

def loadSurface(filename):
    return pygame.image.load(filename)

def loadArray(filename):
    return surfaceToArray(loadSurface(filename))

def surfaceToArray(surface):
    return surfarray.array3d(surface)[:,:,0].astype(N.Float) * (1/255)

def arrayToSurface(array):
    surface = pygame.Surface(array.shape, 0, 24)
    surfarray.blit_array(surface, (N.absolute(array) * 255).astype(N.UnsignedInt8) * 65793)
    return surface

def main():
    pass 

# If Python says run, ...
if __name__ == '__main__':
    # ... run!
    main()












