#!/usr/bin/env python
#
#  pyHatching -- Apply a hatching effect to an image.
#  Copyright (C) 2001 Fabio Dias Moreira
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
## Future directives #################################################

from __future__ import nested_scopes
from __future__ import generators
from __future__ import division

## Imports ###########################################################

import random
from math import *
import bisect

## pyGame -- just import things necessary for image handling
import pygame
import pygame.image
import Numeric as N
import pygame.surfarray as surfarray
from pygame.locals import *

from imageHandling import *

## Constants #########################################################

CORRECTION = 1.0
BLUR_ORDER = 8
JIGGLENESS = 4
EDGE_TOLERANCE = 0.1

## Globals ###########################################################

strokes = {}

## Classes ###########################################################

## Functions #########################################################

def wrap(value, min, max):
    if value < min:
        return value + max - min
    elif value > max:
        return value - max + min
    return value

def thresholdImage(image, level):
    temp = N.sign(image - level)
    return (temp + abs(temp)) * (1/2)

def putStroke(x, y, angle, image):
    imageWidth, imageHeight = image.get_rect().size
    availableAngles = strokes.keys()
    availableAngles.sort()
    angle1 = availableAngles[bisect.bisect_left(availableAngles, angle - 1e-5)]
    angle2 = availableAngles[bisect.bisect_right(availableAngles, angle - 1e-5)]
    if abs(angle - angle1) <= abs(angle - angle2):
        stroke = strokes[angle1]
    else:
        stroke = strokes[angle2]
    strokeWidth, strokeHeight = stroke.get_rect().size
    top, left = x - strokeWidth//2, y - strokeHeight//2
    image.blit(stroke, [[top, left], stroke.get_rect().size])

def initStrokes():
    global strokes
    for i in range(0, 180+1, 5):
        strokes[i] = loadSurface('strokes/stroke-' + `i` + '.png')
    strokes[181] = strokes[180]

def main():
    print \
"""pyHatching version 1.0, Copyright (C) 2001 Fabio Dias Moreira
pyHatching comes with ABSOLUTELY NO WARRANTY; for details see LICENSE.txt.
This is free software, and you are welcome to redistribute it
under certain conditions; see LICENSE.txt for details."""
    pygame.init()
    #print 'Press the mouse button to advance image.'
    filename = raw_input("Filename of the image to filter: ")
    print 'Press the "s" key to save the current image.'
    screen = pygame.display.set_mode([100, 100], 0, 32)
    source = loadArray(filename)
    sourceShape = source.shape
    sourceWidth, sourceHeight = sourceShape
    initStrokes()
    print 'Finding stroke directions...'
    print '  Finding gradients...',
    gradX = circularConvolution(source, \
                                N.array([[-1, 0, 1]]), (1, 1))
    gradY = circularConvolution(source, \
                                N.array([[-1], [0], [1]]), (1, 1))
    absGrad = N.maximum(N.ones(source.shape) * 1e-05, N.sqrt(gradX * gradX + gradY * gradY))
    print 'Done.'
    print '  Finding angles and perpendiculars...',
    # Find the perpendicular angle
    dirGrad = N.arccos(N.divide_safe(gradY, absGrad)) * (180 / pi) - 90
    # Now we have angles ranging from -90..90. Add 180 to the negative angles.
    temp = N.sign(dirGrad)
    dirGrad = dirGrad + (temp - abs(temp)) * -90
    print 'Done.'
    print '  Blurring directions...',
    dirGrad = blurImage(dirGrad, BLUR_ORDER)[BLUR_ORDER:-BLUR_ORDER, \
                                             BLUR_ORDER:-BLUR_ORDER]
    print 'Done.'
    print 'Done.'

    print 'Dithering source image...',
    source = source[BLUR_ORDER:-BLUR_ORDER, BLUR_ORDER:-BLUR_ORDER]
    sourceShape = source.shape
    sourceWidth, sourceHeight = source.shape[:2]
    base = N.zeros(source.shape, N.Float)
    ditherMatrix = (N.array([[ 1,  9,  3, 11], \
                             [13,  5, 15,  7], \
                             [ 4, 12,  2, 10], \
                             [16,  8, 14,  6]]) - 0.5) * (1/16)
    for i in range(4):
        for j in range(4):
            base[i::4, j::4] = thresholdImage(1 - source[i::4, j::4], \
                                              ditherMatrix[i, j])
    print 'Done.'

    print 'Picking stroke points...',
    
    sourceArea = sourceWidth * sourceHeight
    sourceIntensity = N.sum(N.sum(source))
    tempStroke = pygame.Surface(strokes[0].get_rect().size, 0, 32)
    tempStroke.fill((255, 255, 255))
    tempStroke.blit(strokes[0], [[0, 0], strokes[0].get_rect().size])
    tempStroke = surfaceToArray(tempStroke)
    numSamplePoints = sourceArea / N.sum(N.sum(1 - tempStroke)) * CORRECTION
    ratio = sourceArea / numSamplePoints
    spacingX = sqrt(ratio)
    spacingY = sqrt(ratio)
    sampleBase = N.zeros(sourceShape, N.Float)
    strokeInfo = []
    for i in range((spacingX - sourceWidth % spacingX) // 2, \
                   sourceWidth,
                   spacingX):
        for j in range((spacingY - sourceHeight % spacingY) // 2, \
                       sourceHeight,
                       spacingY):
            realI = wrap(i + random.randrange(-JIGGLENESS, JIGGLENESS), 0, sourceWidth - 1)
            realJ = wrap(j + random.randrange(-JIGGLENESS, JIGGLENESS), 0, sourceHeight - 1)
            if base[realI, realJ]:
                strokeInfo.append((realI, realJ, dirGrad[realI, realJ]))
    print 'Done.'

    print 'Drawing strokes...',
    finalImage = pygame.Surface(sourceShape)
    finalImage.fill((255, 255, 255))
    for info in strokeInfo:
        putStroke(info[0], info[1], info[2], finalImage)
    print 'Done.'

    temp = thresholdImage(blurImage(absGrad, BLUR_ORDER // 4), EDGE_TOLERANCE)
##    showArray(dirGrad, 'xc', 1/180)
    finalImage = surfaceToArray(finalImage)
    finalImage = normalizeImage(finalImage + blurImage(thresholdImage(blurImage(temp, 3 * BLUR_ORDER // 8), 0.01), 3 * BLUR_ORDER // 8)[BLUR_ORDER:-BLUR_ORDER, BLUR_ORDER:-BLUR_ORDER])
    finalImage *= blurImage(thresholdImage(1 - blurImage(temp, BLUR_ORDER // 2), 0.7))[BLUR_ORDER:-BLUR_ORDER, BLUR_ORDER:-BLUR_ORDER]
    showArray(finalImage, 'output')
    #showArray(finalImage * 0.2 + source * 0.8)

# If Python says run, ...
if __name__ == '__main__':
    # ... run!
    main()





